from ._dcca import DCCA
from ._dcca_barlow_twins import BarlowTwins
from ._dcca_ey import DCCA_EY
from ._dcca_gh import DCCA_GH
from ._dcca_svd import DCCA_SVD
from ._dcca_noi import DCCA_NOI
from ._dcca_sdl import DCCA_SDL
from ._dtcca import DTCCA
from ._dgcca import DGCCA
